/*
 *  linux/init/main.c
 *
 *  Copyright (C) 1991, 1992  Linus Torvalds
 *
 *  GK 2/5/95  -  Changed to support mounting root fs via NFS
 *  Added initrd & change_root: Werner Almesberger & Hans Lermen, Feb '96
 *  Moan early if gcc is old, avoiding bogus kernels - Paul Gortmaker, May '96
 */

#define __KERNEL_SYSCALLS__

#include <asm/system.h>  
#include <asm/io.h>
#include <linux/types.h>
#include <linux/sched.h>
#include <linux/kernel.h>
#include <linux/tty.h>
#include <linux/head.h>
#include <linux/unistd.h>
#include <linux/timer.h>
#include <linux/utsname.h>
#include <linux/ioport.h>
#include <linux/mm.h>
#include <linux/major.h>
#include <asm/bugs.h>



/*
 * Versions of gcc older than that listed below may actually compile
 * and link okay, but the end product can have subtle run time bugs.
 * To avoid associated bogus bug reports, we flatly refuse to compile
 * with a gcc that is known to be too old from the very beginning.
 */
#if __GNUC__ < 2 || (__GNUC__ == 2 && __GNUC_MINOR__ < 6)
#error sorry, your GCC is too old. It builds incorrect kernels.
#endif

static char printbuf[1024];


static int init(void *);

extern void init_IRQ(void);
extern long console_init(long, long);
extern void time_init(void);

static unsigned long memory_start = 0;
static unsigned long memory_end = 0;


/* add by G.F. */
extern int kbd_init(void);


/* this should be approx 2 Bo*oMips to start (note initial shift), and will
   still work even if initially too large, it will just take slightly longer */
unsigned long loops_per_sec = (1<<12);

int root_mountflags = MS_RDONLY;


/* add by G.F. */
extern void smlnj_main(void);

/*
 *	Activate the first processor.
 */
asmlinkage void start_kernel(void)
{
	char * command_line;
/*
 * Interrupts are still disabled. Do necessary setups, then
 * enable them
 */
	setup_arch(&command_line, &memory_start, &memory_end);
	memory_start = paging_init(memory_start,memory_end);
	trap_init();    
	init_IRQ();

	/* two following functions needed to setup timer - ZK */

	sched_init();
	time_init();

	memory_start = console_init(memory_start,memory_end);

	memory_start = kmalloc_init(memory_start,memory_end);


	mem_init(memory_start,memory_end);
	printk("memory = %ld ... %ld (0x%lx ... 0x%lx)\n",
	       memory_start, memory_end, memory_start,
		memory_end);


	/* add by G.F.*/
	kbd_init();
	smlnj_main();
	printk("after smlnj_main, rebooting .....\n");
	while(1); 

}
