/* main.c
 *
 * COPYRIGHT (c) 1993 by AT&T Bell Laboratories.
 *
 * This is the main routine for the interactive version of SML/NJ.
 */

#ifndef F2A_RTS /* G.F.*/
#include <ctype.h>
#else 
#include <linux/ctype.h>
#include "../gc/heap.h"
#endif 

#include "ml-base.h"
#include "ml-options.h"
#include "ml-limits.h"

#ifdef COLLECT_STATS
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include "stats-data.h"
#endif

#ifdef OLD_RTS /* G.F.*/
FILE		*DebugF = NULL;

#ifdef TARGET_BYTECODE
FILE		*BC_stdout = NULL;
#endif
#endif /* G.F.*/

/* Runtime globals */
bool_t		SilentLoad = TRUE;
bool_t          DumpObjectStrings = FALSE;
bool_t		GCMessages = FALSE;
char		**RawArgs;
#ifdef F2A_RTS
char		*MLCmdName = "hello-ml";
char		dummy_space [4] = {'\0', '\0', '\0', '\0'};
char		**CmdLineArgs = (char **) dummy_space;
#else
char		**CmdLineArgs;	/* does not include the command name (argv[0]) */
char		*MLCmdName;	/* the command name used to invoke the runtime */
#endif /* F2A_RTS */
#ifdef HACKED_STANDALONE
bool_t          StandAlone = TRUE;
#endif

/* local variables */
PVT bool_t	isBoot = FALSE;	/* true if we should bootstrap a system */
PVT char	*LoadImage	/* the path name of the image file to load */
		    = DFLT_IMAGE;
PVT char	*BootFrom	/* the boot source (bin file directory). */
		    = NULL;
PVT int		NumProcs = 1;	/* not used */

#ifdef OLD_RTS /* G.F.*/
PVT void ParseOptions (int argc, char **argv, heap_params_t **heapParams);
#endif  /* G.F. */

#ifdef F2A_RTS /*G.F.*/
void smlnj_main (void)
#endif /*G.F.*/
#ifdef OLD_RTS /*G.F.*/
main(int agrc, char *argv[])
#endif /* G.F.*/ 
{
    heap_params_t	*heapParams;

#ifndef F2A_RTS /* G.F.*/
    DebugF = stderr;
  /* process the command-line options */
    ParseOptions (argc, argv, &heapParams);
#else /* copy from ParseHeapParams() in init-gc.c */ 
    if ((heapParams = NEW_OBJ(heap_params_t)) == NIL(heap_params_t *)) {
	Die("unable to allocate heap_params");
    }
  /* We use 0 or "-1" to signify that the default value should be used. */
    heapParams->allocSz = 0;
    heapParams->numGens = -1;
    heapParams->cacheGen = -1;
    
#endif /* G.F.*/

#ifdef TARGET_BYTECODE
    if (BC_stdout == NULL)
	BC_stdout = fdopen(dup(1), "w");
#endif

    InitTimers ();
    RecordGlobals ();
    InitCFunList ();

#ifdef MP_SUPPORT
    MP_Init();
#endif

  /* start ML */
    if (isBoot) {
#ifdef FULL_SML /* G.F. */
	BootML (BootFrom, heapParams);
#endif  /* G.F. */
    }
    else { /* load an image */
	LoadML (LoadImage, heapParams);
    }

    Exit (0);

} /* end of main. */

#ifdef OLD_RTS /* G.F.*/
/* ParseOptions:
 *
 * Parse the command-line options.
 */
PVT void ParseOptions (int argc, char **argv, heap_params_t **heapParams)
{
    char	option[MAX_OPT_LEN], *optionArg, **nextArg;
    bool_t	errFlg = FALSE;

  /* first scan for any heap/GC parameters */
    if ((*heapParams = ParseHeapParams(argv)) == NIL(heap_params_t *))
	errFlg = TRUE;

    RawArgs = argv;
    CmdLineArgs = NEW_VEC(char *, argc);
    MLCmdName = *argv++;
#ifdef HACKED_STANDALONE
    LoadImage = MLCmdName;
#endif
    nextArg = CmdLineArgs;
    while (--argc > 0) {
	char	*arg = *argv++;

#define MATCH(opt)	(strcmp(opt, option) == 0)
#define CHECK(opt)	{						\
	if (optionArg[0] == '\0') {					\
	    errFlg = TRUE;						\
	    Error("missing argument for \"%s\" option\n", opt);		\
	    continue;							\
	}								\
    } /* CHECK */

	if (isRuntimeOption(arg, option, &optionArg)) {
	    if (MATCH("boot")) {
		CHECK("boot");
		isBoot = TRUE;
		BootFrom = optionArg;
	    }
	    else if (MATCH("load")) {
		CHECK("load");
		LoadImage = optionArg;
#ifdef HACKED_STANDALONE
		LoadImage = MLCmdName;
#endif
	    }
	    else if (MATCH("cmdname")) {
		CHECK("cmdname");
		MLCmdName = optionArg;
#ifdef HACKED_STANDALONE
		StandAlone = FALSE;
#endif
	    }
#ifdef MP_SUPPORT
	    else if (MATCH("nprocs")) {
		CHECK("nprocs");
		NumProcs = atoi(optionArg);
		if (NumProcs < 0)
		    NumProcs = 0;
		else if (NumProcs > MAX_NUM_PROCS)
		    NumProcs = MAX_NUM_PROCS;
	    }
#endif
	    else if (MATCH("quiet")) {
		SilentLoad = TRUE;
	    }
	    else if (MATCH("verbose")) {
		SilentLoad = FALSE;
	    }
	    else if (MATCH("objects")) {
	      DumpObjectStrings = TRUE;
	    }
	    else if (MATCH("debug")) {
		CHECK("debug");
		if ((DebugF = fopen(optionArg, "w")) == NULL) {
		    DebugF = stderr; /* restore the file pointer */
		    errFlg = TRUE;
		    Error("unable to open debug output file \"%s\"\n", *(argv[-1]));
		    continue;
		}
	    }
#ifdef COLLECT_STATS
	    else if (MATCH("stats")) {
		CHECK("stats");
		StatsFD = open (optionArg, O_WRONLY|O_TRUNC|O_CREAT, 0666);
		if (StatsFD == -1) {
		    errFlg = TRUE;
		    Error("unable to open statistics file \"%s\"\n", *(argv[-1]));
		    continue;
		}
	    }
#endif
#ifdef TARGET_BYTECODE
	    else if (MATCH("dump")) {
		CHECK("dump");
		BC_stdout = fopen (optionArg, "w");
	    }
#ifdef INSTR_TRACE
	    else if (MATCH("trace")) {
		extern bool_t	traceOn;
		traceOn = TRUE;
	    }
#endif
#endif
	}
	else {
	    *nextArg++ = arg;
	}
    } /* end while */

    *nextArg = NIL(char *);

    if (errFlg)
	Exit (1);

} /* end of ParseOptions */
#endif /* G.F.*/

/* Exit:
 * Exit from the ML system.
 */
void Exit (int code)
{
int i, j;

#if COUNT_REG_MASKS
    DumpMasks();
#endif
#ifdef COLLECT_STATS
    if (StatsFD >= 0) {
	STATS_FLUSH_BUF();
	close (StatsFD);
    }
#endif
#if (defined(TARGET_BYTECODE) && defined(INSTR_COUNT))
    {
	extern void PrintInstrCount (FILE *f);
	PrintInstrCount (DebugF);
    }
#endif
#ifdef OLD_RTS /* G.F.*/
    exit (code);
#endif

#ifdef F2A_RTS /* G.F.*/
printk("Exit: Dying ....\n");
#endif /* G.F.*/
} /* end of Exit */

