/* interrupt.c
 *
 * specific runtime code to support ML interrupt handler.
 */

#include "ml-unixdep.h"
#include "ml-base.h"
#include "ml-limits.h"
#include "ml-state.h"
#include "vproc-state.h"
#include "ml-objects.h"
#include "ml-globals.h"
#include "ml-interrupt.h"
#include "signal-sysdep.h"
#include <asm/system.h>		/* cli() sti() */
#include <linux/ptrace.h>

#include <asm/io.h>  /* inw, outw */
#include <stdio.h>  /* eixt() */

#ifdef USE_ZERO_LIMIT_PTR_FN
Addr_t		SavedPC;
extern		ZeroLimitPtr[];
#endif

#define SELF_VPROC        (VProc[0])

/* CIntHandler:
 *
 * The C interrupt handler for interrupts that are to be passed to the ML handler.
 */

#define TIMEGAP 1

int timerInterrupts = 0;	/* for use by gc to check time */

void CIntHandler (int irq, void* dev_id, struct pt_regs * regs)
{
    SigContext_t    sc;
    vproc_state_t   *vsp = SELF_VPROC;

static int timer100=0; 

    cli();	/* mask interrupts, clear IF bit in EFLAG */		
    if (irq==0){
      timerInterrupts++;
      timer100++;
      if (timer100==TIMEGAP){
	timer100=0; }
      else { 
	sti(); 
	return;
      }
    }	

    EnqueueInterrupt (vsp, irq);

    if (vsp->vp_inMLFlag) {
	CSigHandler(SIG_INTERRUPT, sc); 
    }
    else; /* printk("!!!not inML!!!\n"); */
    sti(); 	/* unmask interrupts, set IF bit in EFLAG*/

} /* end of CIntHandler */


#define MAX_INTERRUPTS 32
static pending_int_t interrupts [MAX_INTERRUPTS];
static int free_interrupts = 0xffffffff;


static pending_int_t * alloc_int ()
{
  int i;
  for (i = 0; i < MAX_INTERRUPTS; i++) {
    if (((free_interrupts >> i) & 0x1) == 1) {
      free_interrupts &= ~(1 << i);
      return interrupts + i;
    }
  }
  return NULL;
}


static void free_int (pending_int_t * interrupt)
{
  int i = interrupt - interrupts;
  if ((i >= 0) && (i < MAX_INTERRUPTS)) {
    free_interrupts |= (1 << i);
  }
  else{ 
        Die("NOT A LEGAL INTERRUPTS ELEMENT");
  }
}

/* add an interrupt to the pending queue */
void EnqueueInterrupt(vproc_state_t *vsp, int IntID){
	pending_int_t* newInt;
  	newInt = alloc_int();
  	if (newInt == NULL) {
		Die("EnqueueInterrupt: run out of interrupts\n");
	}
	if(vsp->vp_numPendingInts){
            	(vsp->vp_endPendingIntQ)->next = newInt;
	    	vsp->vp_endPendingIntQ = (vsp->vp_endPendingIntQ)->next;
	}	
	else {
		vsp->vp_pendingIntQ = vsp->vp_endPendingIntQ =newInt;
	}
	vsp->vp_numPendingInts++;
	newInt->IntNum = IntID;
	newInt->next = NULL;
}

        
/* delete an interrupt from the pending queue */ 
void DequeueInterrupt(vproc_state_t *vsp){
  	pending_int_t* oldInt;
	if (vsp->vp_numPendingInts){
		oldInt = vsp->vp_pendingIntQ;
	 	vsp->vp_pendingIntQ = oldInt->next;
	 	free_int(oldInt);
		vsp->vp_numPendingInts--;
	} 
	if (!vsp->vp_numPendingInts)
	   vsp->vp_endPendingIntQ = vsp->vp_pendingIntQ = NULL;
}


bool_t IsEmpty(vproc_state_t *vsp) {
	 if((vsp!=NULL) && (vsp->vp_numPendingInts!=0) && (vsp->vp_pendingIntQ != NULL))
		return FALSE;
	 else return TRUE;
}
	
int ChooseInterrupt(ml_state_t *msp) {
    int ID;  
    vproc_state_t *vsp = msp->ml_vproc;	 
    cli();
    if (IsEmpty(vsp)) {
        sti();
	return -1;
    }	
    ID = vsp->vp_pendingIntQ->IntNum;
    DequeueInterrupt(vsp);
    sti();	
    return  ID;
}

