(* signature of the device driver for the network adaptor *)
(* translated to SML and adapted from Linux Debian version 2.0.33
   files:
          linux/if.h

   GF 2/1999
 *)

signature NETDEVCONSTANT = sig
	(* following constant value came from linux/if.h *)
	(* Standard interface flags. *)
	val	IFF_UP		:word	(* interface is up		*)
	val	IFF_BROADCAST	:word		(* broadcast address valid	*)
	val	IFF_DEBUG	:word		(* turn on debugging		*)
	val	IFF_LOOPBACK	:word		(* is a loopback net		*)
	val	IFF_POINTOPOINT	:word		(* interface is has p-p link	*)
	val	IFF_NOTRAILERS	:word		(* avoid use of trailers	*)
	val	IFF_RUNNING	:word		(* resources allocated		*)
	val	IFF_NOARP	:word		(* no ARP protocol		*)
	val	IFF_PROMISC	:word		(* receive all packets		*)
	(* Not supported *)
	val	IFF_ALLMULTI	:word		(* receive all multicast packets*)

	val IFF_MASTER		:word		(* master of a load balancer 	*)
	val IFF_SLAVE		:word		(* slave of a load balancer	*)

	val IFF_MULTICAST	:word		(* Supports multicast		*)
	val IFF_SOFTHEADERS 	:word		(* Device cannot construct headers
						 * until broadcast time. Therefore
						 * SOCK_PACKET must call header
						 * construction. Private flag.
						 * Never visible outside of kernel.
						 *)


	(* following constant came from linux/if_arp.h *)
	(* ARP protocol HARDWARE identifiers. *)
	val ARPHRD_NETROM	:int 		(* from KA9Q: NET/ROM pseudo	*)
	val ARPHRD_ETHER  	:int	(* Ethernet 10Mbps		*)
	val ARPHRD_EETHER 	:int	(* Experimental Ethernet	*)
	val ARPHRD_AX25   	:int	(* AX.25 Level 2		*)
	val ARPHRD_PRONET 	:int	(* PROnet token ring		*)
	val ARPHRD_CHAOS  	:int	(* Chaosnet			*)
	val ARPHRD_IEEE802	:int	(* IEEE 802.2 Ethernet/TR/TB	*)
	val ARPHRD_ARCNET 	:int	(* ARCnet			*)
	val ARPHRD_APPLETLK	:int	(* APPLEtalk			*)
	val ARPHRD_DLCI	  	:int	(* Frame Relay DLCI		*)
	val ARPHRD_METRICOM	:int	(* Metricom STRIP (new IANA id)	*)

	(* Dummy types for non ARP hardware *)
	val ARPHRD_SLIP	   :int
	val ARPHRD_CSLIP   :int
	val ARPHRD_SLIP6   :int
	val ARPHRD_CSLIP6  :int
	val ARPHRD_RSRVD   :int		(* Notional KISS type 		*)
	val ARPHRD_ADAPT   :int
	val ARPHRD_PPP	   :int

	val ARPHRD_TUNNEL  	:int		(* IPIP tunnel			*)
	val ARPHRD_TUNNEL6 	:int		(* IPIP6 tunnel			*)
	val ARPHRD_FRAD	   	:int		(* Frame Relay Access Device	*)
	val ARPHRD_SKIP	   	:int		(* SKIP vif			*)
	val ARPHRD_LOOPBACK	:int	(* Loopback device		*)
	val ARPHRD_LOCALTLK	:int	(* Localtalk device		*)
	val ARPHRD_FDDI	   	:int	(* Fiber Distributed Data Interface *)

	(* ARP protocol opcodes. *)
	val ARPOP_REQUEST  	:int	(* ARP request			*)
	val ARPOP_REPLY    	:int	(* ARP reply			*)
	val ARPOP_RREQUEST 	:int	(* RARP request			*)
	val ARPOP_RREPLY   	:int	(* RARP reply			*)

	(* following constant came from linux/socket.h *)
	(* Supported address families. *)
	val AF_UNSPEC	:int
	val AF_UNIX	:int	(* Unix domain sockets 		*)
	val AF_INET	:int	(* Internet IP Protocol 	*)
	val AF_AX25	:int	(* Amateur Radio AX.25 		*)
	val AF_IPX	:int	(* Novell IPX 			*)
	val AF_APPLETALK :int	(* Appletalk DDP 		*)
	val AF_NETROM	:int	(* Amateur radio NetROM 	*)
	val AF_BRIDGE	:int	(* Multiprotocol bridge 	*)
	val AF_AAL5	:int	(* Reserved for Werner's ATM 	*)
	val AF_X25	:int	(* Reserved for X.25 project 	*)

	(* Protocol families, same as address families. *)
	val PF_UNSPEC	:int
	val PF_UNIX	:int
	val PF_INET	:int
	val PF_AX25	:int
	val PF_IPX	:int
	val PF_APPLETALK :int
	val PF_NETROM	:int
	val PF_BRIDGE	:int
	val PF_AAL5	:int
	val PF_X25	:int

end
