
(*

	FoxNet: The Fox Project's Communication Protocol Implementation Effort
	Brian Milnes (Brian.Milnes@cs.cmu.edu)
	Nick Haines  (Nick.Haines@cs.cmu.edu)
	Edoardo Biagioni <Edoardo.Biagioni@cs.cmu.edu>
        Ken Cline    (Ken.Cline@cs.cmu.edu)
	Fox Project
	School of Computer Science
	Carnegie Mellon University
	Pittsburgh, Pa 15139-3891

		i.	Abstract

	This module contains functions for converting data
        between strings, Bytearrays, and ubyte1 flat_vectors.

	Additionally, the function "flatten" views any fully
        flat (i.e. containing no boxed items) locative as a
        ubyte1 flat_vector locative.

	"Big Oh" algorithmic performance for the current
	implementation is given in comments for each function.

		ii.	Table of Contents

	i.	Abstract
	ii.	Table of Contents
	iii.	RCS Log
	1.	signature CONVERT


		iii.	RCS Log
	
$Log: convert.sig,v $
Revision 1.4  1993/12/21  15:06:15  cline
added comments.

Revision 1.3  1993/12/21  14:11:21  cline
Eliminated opens.

Revision 1.2  1993/12/15  15:25:14  cline
Removed #s from changelog.

Revision 1.1  1993/10/20  14:24:34  cline
Initial revision


		1.	signature CONVERT
*)

local
  structure BA = ByteArray
  structure FV = Flat_Vector
  structure L  = Locative
in
  signature CONVERT =
  sig

    (* raised when flatten is called on an object with boxed components *)
    exception flatten_box

    val string2flat_vector (*O(n)*)     : string -> ubyte1 FV.flat_vector
    val flat_vector2string (*O(n^2)*)   : ubyte1 FV.flat_vector -> string


    val bytearray2flat_vector (*O(n)*) : BA.bytearray -> ubyte1 FV.flat_vector
    val flat_vector2bytearray (*O(n)*) : ubyte1 FV.flat_vector -> BA.bytearray


    (*
       flatten takes any flat data and flattens it into a vector of
       bytes. This can be done with forge, but flatten is type safe.

       flatten runs in time proprotional to the size of the descriptor
       of its argument.
     *)
    val flatten : 'a locative -> ubyte1 FV.flat_vector locative
  end (* signature CONVERT *)
end (* local *)