(*

	FoxNet: The Fox Project's Communication Protocol Implementation Effort
	Edoardo Biagioni (Edoardo.Biagioni@cs.cmu.edu)
	Brian Milnes (Brian.Milnes@cs.cmu.edu)
	Ken Cline (Kenneth.Cline@cs.cmu.edu)
        Nick Haines (Nick.Haines@cs.cmu.edu)
	Fox Project
	School of Computer Science
	Carnegie Mellon University
	Pittsburgh, Pa 15139-3891

		i.	Abstract



		ii.	Table of Contents

	i.	Abstract
	ii.	Table of Contents
	iii.	RCS Log
	1.	functor Krb_Protocol

		iii.	RCS Log
	
$Log: krbproto.fun,v $
Revision 1.2  1994/05/24  14:33:30  robby
 Ken's restructuring following a code review.

Revision 1.1  1993/12/14  21:51:48  cline
Initial revision


		1.	functor Krb_Protocol
*)

(* see /usr/src/usr/misc/.kerberos/src/include/prot.h *)

functor Krb_Protocol(structure Realm : KRB_REALM) : KRB_PROTOCOL =
struct

  exception Message_Type
  exception Error_Code

  type name     = string
  type instance = string
  type realm    = string

  datatype ticket = Ticket of ubyte1 Flat_Vector.flat_vector
  datatype key    = Key    of ubyte1 Flat_Vector.flat_vector
  datatype principal =
             Principal       of {name:string, instance:string, realm:string}
  datatype local_principal =
             Local_Principal of {name:string, instance:string}

  (*
    Credentials contain all of the information returned from Kerberos
    when a ticket is granted.  As a rule, clients will treat credentials
    as opaque types.
   *)
  datatype
    credential =
      Credential of {service:principal,
                     session_key:key,
                     issue_date:ubyte4,
                     kvno:ubyte1,		(* key version number *)
                     lifetime:ubyte1,
                     ticket:ticket}          

  (*
    According to RFC 1340, Kerberos should use port 88.  However,
    750 is the port number found in the C code for the CMU version.
    Additionally, CMU Kerberos talks to MIT Kerberos (which therefore
    must also recognize port 750.
   *)
  val port = 2u750  (* not all systems have /etc/services *)

  val protocol_version = 1u4

  val byte_order =
    let val test = ByteArray.array (2,0)
    in
      Byte2.update (test, 0, 2u1);
      Byte1.sub (test, 0)
    end


  val ticket_granting_service =
      Local_Principal {name="krbtgt", instance=Realm.local_realm ()}

  (* Kerberos message types *)
  datatype auth_message_type =
    Kdc_Request | Kdc_Reply | Appl_Request | Appl_Request_Mutual | Err_Reply
  | Private | Safe | Appl_Err | Die

  (* Note: lsb holds byte order information. *)
  fun auth_message_type2byte a =
    Byte1.|| (byte_order,
              case a of Kdc_Request => 1u2
              | Kdc_Reply           => 1u4
              | Appl_Request        => 1u6
              | Appl_Request_Mutual => 1u8
              | Err_Reply           => 1u10
              | Private             => 1u12
              | Safe                => 1u14
              | Appl_Err            => 1u16
              | Die                 => 1u126)

  fun byte2auth_message_type b =
    case Byte1.&& (b, 1uxfe) of
      1u2   => Kdc_Request
    | 1u4   => Kdc_Reply
    | 1u6   => Appl_Request
    | 1u8   => Appl_Request_Mutual
    | 1u10  => Err_Reply 
    | 1u12  => Private
    | 1u14  => Safe
    | 1u16  => Appl_Err 
    | 1u126 => Die
    | _ => raise Message_Type


  (* Kerberos error codes *)
  datatype kerb_error_code =
    Ok | Name_Exp | Service_Exp | Auth_Exp | Pkt_Ver | Name_Mast_Key_Ver
  | Serv_Mast_Key_Ver | Byte_Order | Principal_Unknown | Principal_Not_Unique
  | Null_Key

  fun kerb_error_code2byte Ok                   = 1u00
    | kerb_error_code2byte Name_Exp             = 1u01
    | kerb_error_code2byte Service_Exp          = 1u02
    | kerb_error_code2byte Auth_Exp             = 1u03
    | kerb_error_code2byte Pkt_Ver              = 1u04
    | kerb_error_code2byte Name_Mast_Key_Ver    = 1u05
    | kerb_error_code2byte Serv_Mast_Key_Ver    = 1u06
    | kerb_error_code2byte Byte_Order           = 1u07
    | kerb_error_code2byte Principal_Unknown    = 1u08
    | kerb_error_code2byte Principal_Not_Unique = 1u09
    | kerb_error_code2byte Null_Key             = 1u10

  fun byte2kerb_error_code 1u00 = Ok
    | byte2kerb_error_code 1u01 = Name_Exp
    | byte2kerb_error_code 1u02 = Service_Exp
    | byte2kerb_error_code 1u03 = Auth_Exp
    | byte2kerb_error_code 1u04 = Pkt_Ver
    | byte2kerb_error_code 1u05 = Name_Mast_Key_Ver
    | byte2kerb_error_code 1u06 = Serv_Mast_Key_Ver
    | byte2kerb_error_code 1u07 = Byte_Order
    | byte2kerb_error_code 1u08 = Principal_Unknown
    | byte2kerb_error_code 1u09 = Principal_Not_Unique
    | byte2kerb_error_code 1u10 = Null_Key
    | byte2kerb_error_code _ = raise Error_Code

end
