(*

	FoxNet: The Fox Project's Communication Protocol Implementation Effort
	Edoardo Biagioni (Edoardo.Biagioni@cs.cmu.edu)
	Brian Milnes (Brian.Milnes@cs.cmu.edu)
	Ken Cline (Kenneth.Cline@cs.cmu.edu)
        Nick Haines (Nick.Haines@cs.cmu.edu)
	Fox Project
	School of Computer Science
	Carnegie Mellon University
	Pittsburgh, Pa 15139-3891

		i.	Abstract

	This module contains declarations relating to the
	Kerberos protocol, including datatypes and constants.

		ii.	Table of Contents

	i.	Abstract
	ii.	Table of Contents
	iii.	RCS Log
	1.	signature KRB_PROTOCOL

		iii.	RCS Log
	
$Log: krbproto.sig,v $
Revision 1.2  1994/05/24  14:33:32  robby
 Ken's restructuring following a code review.

Revision 1.1  1993/12/14  21:51:32  cline
Initial revision


		1.	signature KRB_PROTOCOL
*)

signature KRB_PROTOCOL =
sig

  exception Message_Type
  exception Error_Code

  type name
  type instance
  type realm

  datatype ticket = Ticket of ubyte1 flat_vector
  datatype key    = Key    of ubyte1 Flat_Vector.flat_vector
  datatype principal =
             Principal       of {name:name, instance:instance, realm:realm}
  datatype local_principal =
             Local_Principal of {name:name, instance:instance}

  (*
    Credentials contain all of the information returned from Kerberos
    when a ticket is granted.  As a rule, clients will treat credentials
    as opaque types.
   *)
  datatype
    credential =
      Credential of {service:principal,
                     session_key:key,
                     issue_date:ubyte4,
                     kvno:ubyte1,		(* key version number *)
                     lifetime:ubyte1,
                     ticket:ticket}          

  val port : ubyte2	(* UDP port number for Kerberos communication. *)

  val protocol_version : ubyte1

  val byte_order : ubyte1 (* least significant bit indicates host byte order *)

  val ticket_granting_service : local_principal


  (* Kerberos message types *)
  datatype auth_message_type =
    Kdc_Request
  | Kdc_Reply
  | Appl_Request
  | Appl_Request_Mutual
  | Err_Reply
  | Private
  | Safe
  | Appl_Err
  | Die

  val auth_message_type2byte : auth_message_type -> ubyte1

  (* byte2auth_message_type may raise Message_Type *)
  val byte2auth_message_type : ubyte1 -> auth_message_type

  (* Kerberos error codes *)
  datatype kerb_error_code =
    Ok                  
  | Name_Exp            
  | Service_Exp         
  | Auth_Exp            
  | Pkt_Ver             
  | Name_Mast_Key_Ver   
  | Serv_Mast_Key_Ver   
  | Byte_Order          
  | Principal_Unknown   
  | Principal_Not_Unique
  | Null_Key

  val kerb_error_code2byte : kerb_error_code -> ubyte1

  (* byte2kerb_error_code may raise Error_Code *)
  val byte2kerb_error_code : ubyte1 -> kerb_error_code

end
