(*

	FoxNet: The Fox Project's Communication Protocol Implementation Effort
	Edoardo Biagioni (esb@cs.cmu.edu)
	Ken Cline    (Ken.Cline@cs.cmu.edu)
	Nick Haines  (Nick.Haines@cs.cmu.edu)
	Brian Milnes (Brian.Milnes@cs.cmu.edu)
	Fox Project
	School of Computer Science
	Carnegie Mellon University
	Pittsburgh, Pa 15139-3891

		i.	Abstract

	packetfilter.fun: an interface to the OSF1 packetfilter pseudo-device.



		ii.	Table of Contents

	i.	Abstract
	ii.	Table of Contents
	iii.	RCS Log
	1.	functor Packet_Filter
	2.	types and exceptions
	3.	module Trace
	4.	c functions
	5.	internal functions raise_fun and handle_fun
	6.	fun pfopen
	7.	fun close
	8.	fun get_ethernet_address
	9.	fun set_filter
	10.	fun readi.
	11.	fun write
	12.	fun writev

*)

(*

		iii.	RCS Log

$Log:	packetfilter.fun,v $
Revision 1.22  97/11/19  13:50:11  cline
109.32 compatibility

Revision 1.21  97/03/27  13:47:45  esb
removed select.

Revision 1.20  97/03/06  15:45:12  cline
Changed select to use new C function.

Revision 1.19  1997/01/29  19:42:10  cline
changed min to Int.min

Revision 1.18  1996/04/29  17:42:12  esb
replaced "makestring" with "Int.makestring".

Revision 1.17  1996/03/04  20:48:30  derby
Added writev to the packetfilter.

Revision 1.16  1996/02/28  21:12:43  cline
fixed typo (packetfilter.sig -> packetfilter.fun)

Revision 1.15  1996/02/13  15:27:26  cline
removed Word32.W (word32 wrapping hacks)

Revision 1.14  1996/02/01  19:20:59  derby
Added calls to access pty functionality: openptr forkptr and window accessors

Revision 1.13  1996/01/19  23:04:33  esb
adapted to the new wordarray signature.

Revision 1.12  1996/01/15  18:00:49  cline
removed FoxWord

Revision 1.11  1995/11/29  21:55:18  cline
ported to SML/NJ 108.13

Revision 1.10  1995/11/12  16:38:45  esb
adapted to new Word_Array.

Revision 1.9  1995/11/08  23:10:27  cline
*** empty log message ***

Revision 1.8  1995/09/18  19:29:17  esb
first running version.

Revision 1.7  1995/03/12  17:53:08  esb
adapted to new trace.sig.

Revision 1.6  1995/03/07  23:55:43  esb
removed a source of inefficiency.

Revision 1.5  1995/02/21  15:46:37  esb
eliminated debugging statements which were seriously affecting performance.

Revision 1.4  1995/02/04  20:39:46  robby
updated to 107

Revision 1.3  1995/01/14  02:32:01  esb
adapted to new packetfilter.sig

Revision 1.2  1994/11/22  13:58:38  milnes
Removed addressing functor arguments.

Revision 1.1  1994/10/20  17:57:49  cline
Initial revision


*)

(*
	1.	functor Packet_Filter
*)

functor Packet_Filter (structure B: FOX_BASIS
		       val debug_level: int ref option): PACKET_FILTER =
 struct

(*
	2.	types and exceptions
*)

  type T = Posix.FileSys.file_desc
  type filter = Word_Array.T

  exception Packet_Filter

(*
	3.	module Trace
*)

  fun makestring_exn Packet_Filter = SOME "Packet Filter"
    | makestring_exn _ = NONE

  structure Trace = Trace (structure V = B.V
			   val debug_level = debug_level
			   val module_name = "packetfilter.fun"
			   val makestring = makestring_exn)

(*
	4.	internal function handle_fun
*)

  fun handle_fun (x, s) =
       (Trace.print_handled (x, SOME s);
	Trace.print_raise (Packet_Filter, SOME s))

(*
	5.	fun pfopen
*)

  local
   fun make_socket (fd: T): (unit, Socket.dgram) Socket.sock =
        Unsafe.cast fd

   fun make_T (socket: (unit, Socket.dgram) Socket.sock): T =
        Unsafe.cast socket

   fun find_match (string, []) =
        Trace.print_raise (Packet_Filter, SOME "no INET AF")
     | find_match (string, (name, value) :: rest) =
	if string = name then value else find_match (string, rest)

   val inet_af = find_match ("INET", Socket.AF.list ())
(* val sock_packet = find_match ("PACKET", Socket.SOCK.list ()) *)
(* note: sml/nj doesn't define PACKET, which is what we need.  So,
   as a hack, we go into memory and change the numerical value of
   "SEQPACKET" to 10 *)
   val sock_packet = find_match ("SEQPACKET", Socket.SOCK.list ())
   val _ = Unsafe.Word8Array.update (Unsafe.cast sock_packet, 0, 0wx15)

(* another hack: sml/nj doesn't give us a way to create our own addresses,
   so for sending we need to use cast. *)
   val eth0 = map (Word8.fromInt o Char.ord) (String.explode "eth0")
   val eth0vec = Word8Vector.fromList ([0w1, 0w0] (* address family *)
				       @ eth0 @
				       [0w0, 0w0, 0w0, 0w0, 0w0, 0w0,
					0w0, 0w0, 0w0, 0w0, 0w8, 0w0])
   datatype socket_addr = ADDR of Word8Vector.vector
   val send_to = Unsafe.cast (ADDR eth0vec)

   fun htons n =
        if Word16.bigEndian then n
        else
	 let val a = Word8Array.array (2, 0w0)
	 in Word16.update_big (a, 0, Word16.fromInt n);
	    Word16.toInt (Word16.sub (a, 0))
	 end

   val protocol = htons 3

  in

   fun pfopen interface =
        ((make_T (GenericSock.socket' (inet_af, sock_packet, protocol)))
	 handle x => Trace.print_raise_again (x, SOME "pfopen"))

(*
	7.	fun close
*)

   val close = Socket.close o make_socket

(*
	8.	fun get_ethernet_address

	Cheat for now.  Eventually, should do something corresponding to:

  struct ifreq ifr;

  bzero (&ifr, sizeof(ifr));
  strcpy (ifr.ifr_name, "eth0");
  if (ioctl (s, SIOCGIFHWADDR, &ifr) < 0) {
    perror ("ioctl(SIOCGIFHWADDR)");
    exit (1);
  }
  printf ("local address is ");
  for (i = 0; i < 6; i++) {
    printf ("%02x%s", (ifr.ifr_hwaddr.sa_data [i]) & 0xff,
	    (i == 5) ? "" : ":");
  }
  printf (", address family 0x%x%s\n", ifr.ifr_hwaddr.sa_family,
	  (ifr.ifr_hwaddr.sa_family == ARPHRD_ETHER) ? " (ethernet)" : "");


  but for now, I don't seem to have ioctls.
*)

(*
   fun makestring_ethernet buffer =
        let val b0 = Word8Array.sub (buffer, 0)
	    val b1 = Word8Array.sub (buffer, 1)
	    val b2 = Word8Array.sub (buffer, 2)
	    val b3 = Word8Array.sub (buffer, 3)
	    val b4 = Word8Array.sub (buffer, 4)
	    val b5 = Word8Array.sub (buffer, 5)
	    val ts = Word8.toString
        in ts b0^":"^ts b1^":"^ts b2^":"^ts b3^":"^ts b4^":"^ts b5
        end

   fun get_ethernet_address socket =
        let val int_pf = Word32.toInt (Posix.FileSys.fdToWord pf)
	    val buffer = ((c_pfget_address int_pf)
			  handle x => handle_fun (x, "get_ethernet_address"))
	    fun create_array n =
	         if n >= Word8Array.length buffer then NONE
		 else SOME (Word8Array.sub (buffer, n), n + 1)
        in Trace.debug_print (fn _ => "ethernet address (fd = " ^
			      Int.toString int_pf ^ ") is " ^
			      makestring_ethernet buffer);
	   Word_Array.from8 (Word_Array.W8.U_Big.F.new create_array 0)
        end
*)
   fun get_ethernet_address socket =
	let exception Error_In_Ethernet_Address
	    fun create [] = NONE
	      | create (first :: rest) = SOME (first, rest)
	in Word_Array.from8 (Word_Array.W8.U_Big.F.new
			     create [0wx00, 0wx20, 0wxaf, 0wxd6, 0wxa1, 0wx03])
	end
(*
	    fun create_array 0w0 = 0wx00
	      | create_array 0w1 = 0wx20
	      | create_array 0w2 = 0wxaf
	      | create_array 0w3 = 0wxd6
	      | create_array 0w4 = 0wxa1
	      | create_array 0w5 = 0wx03
	      | create_array _ = raise Error_In_Ethernet_Address
	in Word_Array.from8 (Word_Array.W8.U_Big.F.tabulate
			     (create_array, 0w6))
	end
*)

(*
	9.	fun set_filter
*)

(*
  fun join ([], NONE) = NONE
    | join ([], SOME (head, rest)) =
       SOME (head, ([], Word_Array.W8.U_Big.F.next rest))
    | join (head :: rest, other) = SOME (head, (rest, other))

  val zero = Word8.fromInt 0
  fun set_filter (pf, filter) =
       let val filter8 = Word_Array.to8 filter
	   val length = Word8.fromInt (Word.toInt 
				       (Word_Array.W8.U_Big.F.length filter8))
	   (* filter priority - any value greater than 2 should work *)
	   val priority = Word8.fromInt 50
	   val arg = ([priority, length], Word_Array.W8.U_Big.F.next filter8)
	   val array = Word_Array.W8.U_Big.F.new join arg
	   val int_pf = Word32.toInt (Posix.FileSys.fdToWord pf)
	   val (buffer, _, _) = Word_Array.expose (Word_Array.from8 array)
       in Trace.local_print ("setting filter " ^
			     B.Format.makestring (Word_Array.from8 array) ^
			     " for fd = " ^ Int.toString int_pf);
	  ((c_pfset_filter (int_pf, buffer))
	   handle x => handle_fun (x, "set_filter"))
       end
*)
   fun set_filter (socket, filter) = 
        Trace.local_print ("ignoring call to set_filter")

(*
	10.	fun readi.
	Read from packetfilter starting at position i of buf.
*)

   fun readi (socket, buf, count, i) =
        let 
fun pai (a, i, max) =
     if i >= max orelse i >= Word8Array.length a then print "\n"
     else
      (print (Word8.toString (Word8Array.sub (a, i)));
       print (".");
       pai (a, i + 1, max));
(*
fun isbc (a, i, 0) = false
  | isbc (a, i, n) =
     Word8Array.sub (a, i) = 0wxff andalso isbc (a, i + 1, n - 1)
fun ismyarp (a, i) = 
     (Word8Array.sub (a, i) = 0wx8 andalso
      Word8Array.sub (a, i + 1) = 0wx6 andalso
      Word8Array.sub (a, i + 26) = 0wx80 andalso
      Word8Array.sub (a, i + 27) = 0wxAB andalso
      Word8Array.sub (a, i + 28) = 0wxA andalso
      Word8Array.sub (a, i + 29) = 0wx4B) handle _ => false
fun isip (a, i) = Word8Array.sub (a, i) = 0wx8 andalso
                  Word8Array.sub (a, i + 1) = 0wx0
fun isicmp (a, i) = Word8Array.sub (a, i) = 0wx1
*)
	    val (data, first, last) = Word_Array.expose buf
	    val (size, address) =
                    ((Socket.recvArrFrom (make_socket socket,
					  {buf = data,
					   i = i + Word.toInt first}))
		     handle x => handle_fun (x, "readi"))
	in (* print ("packetfilter.fun: received ");
	   pai (data, i + Word.toInt first, 100); *)
	   1514
        end

(*
	11.	fun write
*)

   fun write (socket, buf, n) =
        let val (data, first, last) = Word_Array.expose buf
	    val size = Word.toInt (last - first) + 1
	    val send_data = {buf = data, i = Word.toInt first, sz = SOME size}
        in ((Socket.sendArrTo (make_socket socket, send_to, send_data))
	    handle x => handle_fun (x, "write"));
	   ()
        end

(*
	12.	fun writev

	For now, copy.
*)

(*
  fun writev (pf, bufferList) = 
     let val pf' = Word32.toInt (Posix.FileSys.fdToWord pf)
        fun convert wordArray = 
           let val ( buffer, first, last ) = Word_Array.expose wordArray
           in
              ( buffer, Word.toInt first, Word.toInt (last - first + 0w1) )
           end
     in
        c_writev (pf', map convert bufferList)
     end
*)

   fun writev (socket, buffers) = 
       let fun al (wa, length) =
                let val (buffer, first, last) = Word_Array.expose wa
		in Word.toInt (last - first) + 1 + length
		end
	   val length = B.V.List.fold al buffers 0
	   val buffer = Unsafe.Word8Array.create length
           fun convert (length, []) = ()
	     | convert (index, head :: rest) =
                let val (data, first, last) = Word_Array.expose head
		    val len = Word.toInt (last - first) + 1
		in Word8Array.copy {di = index, dst = buffer, len = SOME len,
				    si = Word.toInt first, src = data};
		   convert (index + len, rest)
		end
	   val send_data = {buf = buffer, i = 0, sz = NONE}
fun pai (a, i, max) =
if i >= max orelse i >= Word8Array.length a then print "\n"
else
 (print (Word8.toString (Word8Array.sub (a, i)));
  print (".");
  pai (a, i + 1, max));
       in convert (0, buffers);
          ((Socket.sendArrTo (make_socket socket, send_to, send_data))
	   handle x => handle_fun (x, "writev"))
	  (* before (print "written: "; pai (buffer, 0, 100)) *)
       end
  end					(* local *)

 end
