(*

        FoxNet: The Fox Project's Communication Protocol Implementation Effort
        Edo Biagioni (esb@cs.cmu.edu)
	Ken Cline (Ken.Cline@cs.cmu.edu)
        Nick Haines (nickh@cs.cmu.edu)
        Brian Milnes (milnes@cs.cmu.edu)
        Fox Project
        School of Computer Science
        Carnegie Mellon University
        Pittsburgh, Pa 15139-3891

	i.	Abstract

	protoextern.fun: external structure for protocols (one possible
	implementation).

	ii.	Table of Contents

	i.	Abstract
	ii.	Table of Contents
	iii.	RCS Log
	1.	functor Protocol_External
	2.	function fold
	3.	functions makestring and makestring_max

	iii.	RCS Log

$Log: protoextern.fun,v $
Revision 1.8  1996/02/23  21:37:17  esb
now allows sub of an empty T, if the length is zero.

Revision 1.7  1996/02/15  19:05:14  esb
makestring_max now prints "..." if some elements are not printed.

Revision 1.6  1996/01/19  23:05:43  esb
adapted to the new wordarray signature.

Revision 1.5  1996/01/15  18:00:49  cline
removed FoxWord

Revision 1.4  1995/11/12  16:40:14  esb
adapted to new Word_Array.

Revision 1.3  1995/07/05  17:44:26  esb
adapted to new wordarray signature.

Revision 1.2  1995/06/23  19:57:05  esb
improved error reporting on exceptions.

Revision 1.1  1995/06/20  17:09:33  esb
Initial revision


	1.	functor Protocol_External
*)

functor Protocol_External (structure B: FOX_BASIS
			   val debug_level: int ref option): EXTERNAL =
 struct

  structure Trace = Trace (structure V = B.V
			   val debug_level = debug_level
			   val module_name = "protoextern.fun"
			   val makestring = fn _ => NONE)

  exception Illegal_Access

  fun illegal_access string = Trace.print_raise (Illegal_Access, SOME string)

(*  GF 09/28/99
  structure WAF = Word_Array.W8.U_Big.F
  structure WAR = Word_Array.W8.U_Big.R
*)

  
(*
  datatype T = Empty
             | Single of Word_Array.W8.T
             | Multiple of Word_Array.W8.T * Word_Array.W8.T list
*)

  type single = {data: Word8Array.array, start: Word.word,
		 length: Word.word}
(*
  datatype T = Empty
             | Single of single
             | Multiple of single * single list
*)

  type T = single list


  fun new array =
      (( if Word8Array.length array < 1 then []
       else  [{ data = array, 
		     start = 0w0, 
		     length = Word.fromInt (Word8Array.length array)}])
	  handle x => Trace.print_raise_again(x, SOME "new") )



(*   GF -- 092799  
  fun new array =
       let val a8 = Word_Array.to8 array
       in if WAF.length a8 < 0w1 then Empty
	  else Single a8
       end
*)

  fun uninitialized size =
    (( if size < 0w1 then []
       else [{data = Word8Array.array(Word.toInt size, 0w0),
	      start = 0w0, length = size }] )
       handle x =>Trace.print_raise_again(x, SOME "uninitialized") )


(*
  fun uninitialized size =
       if size < 0w1 then Empty
       else Single (WAF.create (Word8.fromInt 0, size))
*)

  fun new_offset_size (array, offset, size) =
     ((  [{data = array, start = offset, length = size}])
        handle x =>Trace.print_raise_again(x, SOME "new_offset_size") )

  fun size [] = 0w0
    | size list = 
	((let fun single_size {data, start, length} = length
       in B.V.List.fold (fn (a, remainder_size) =>
		      (single_size a + remainder_size)) list 0w0
       end)
        handle x =>Trace.print_raise_again(x, SOME "size") )

(*
  fun size Empty = 0w0
    | size (Single {length, ...}) = length
    | size (Multiple (head, rest)) =
       B.V.List.fold (fn (a, remainder_size) =>
		      (WAF.length a + remainder_size)) (head :: rest) 0w0
*)


  val empty_sub = { data = Word8Array.array(0, Word8.fromInt 0),
		    start = 0w0,
		    length = 0w0 }
(*
  val empty_sub = Word_Array.from8 (WAF.create (Word8.fromInt 0, 0w0))
*)

  fun sub ([]:T, {sub_start, sub_length}) =
      (( if sub_length = 0w0 then empty_sub else illegal_access "sub of empty T")
       handle x =>Trace.print_raise_again(x, SOME "sub1") )
     |sub ({data, start, length}::[], {sub_start, sub_length}) =
         ( (if (sub_start + sub_length) > length
	  then illegal_access "sub longer than T"
	  else if (sub_start < 0w0 ) orelse (sub_length < 0w0)
	       then illegal_access "illegal sub index or length"
	       else let val a = Word8Array.array(Word.toInt sub_length, 0w0)
		    in  Word8Array.copy{src=data, si=Word.toInt (start+sub_start), 
					len=SOME (Word.toInt sub_length), 
					dst=a, di=0 };
		        List.hd(new a)
		    end	)
         handle x =>Trace.print_raise_again(x, SOME "sub2") )
    | sub (head::tail, {sub_start, sub_length}) =
       ((let val a = Word8Array.array(Word.toInt sub_length, 0w0)    
 	   fun loop([], _, length, _) =
	       if length = 0w0 then ()
	       else illegal_access "sub longer than T"
	     | loop({data, start, length}::tail, 
		    sub_start, sub_length,dst_start) = (
	           if (sub_length + sub_start)> length
	           then(Word8Array.copy{src=data, 
		          		si=Word.toInt(sub_start + start), 
			       	        len=SOME(Word.toInt(length-sub_start)),
					dst=a, di=Word.toInt dst_start };
			 loop(tail, 0w0, sub_length-(length-sub_start),
			      dst_start + (length-sub_start) ))
		    else Word8Array.copy{src=data, 
					 si=Word.toInt(sub_start + start), 
					 len=SOME(Word.toInt(sub_length)),
					 dst=a, di=Word.toInt dst_start } )
       in  loop(head::tail, sub_start, sub_length, 0w0);
	   List.hd( new a)
       end) handle x =>Trace.print_raise_again(x, SOME "sub3") )


      

(*
fun sub (Empty, {start, length}) =
       if length = 0w0 then empty_sub else illegal_access "sub of empty T"
    | sub (Single array, {start, length}) =
       let fun next (0w0, array) = NONE
	     | next (count, array) =
		case WAF.next array of
		   NONE => illegal_access "sub longer than T"
		 | SOME (value, new_array) =>
		    SOME (value, (count - 0w1, new_array))
       in Word_Array.from8
	    (WAF.new next (length, WAF.seek (array, start)))
       end
    | sub (array as (Multiple (head, tail)), {start, length}) =
       let fun next (_, 0w0) = NONE
	     | next ((NONE, []), count) = illegal_access "sub longer than T"
	     | next ((NONE, head :: tail), count) =
	        next ((WAF.next head, tail), count)
	     | next ((SOME (value, head), list), count) =
	        SOME (value, ((WAF.next head, list), count - 0w1))
	   fun skip_loop ([], 0w0) = (NONE, [])
	     | skip_loop ([], _) = illegal_access "sub starts past end of T"
	     | skip_loop (head :: tail, count) =
	        let val head_length = WAF.length head
		in if count >= head_length then
		    skip_loop (tail, count - head_length)
		   else (NONE, WAF.seek (head, count) :: tail)
		end
       in Word_Array.from8 (WAF.new next (skip_loop (head :: tail, start),
					  length))
       end
*)


  fun update ([], _, _) = illegal_access "updating empty T"
    | update ({data,start,length}::[], sub_start, value) =
       ((let val new_l = Word.fromInt (Word8Array.length value)
	   fun loop(old_array, old_i, new_i) = 
	       if new_i = new_l   then()
	       else if old_i = length 
		    then illegal_access "update longer than T"
		    else (
			  Word8Array.update(old_array, 
					    Word.toInt(old_i+start), 
					    Word8Array.sub(value,
							   Word.toInt new_i));
			  loop(old_array, old_i+0w1, new_i+0w1) ) 
       in loop(data, sub_start, 0w0)
       end) handle x =>Trace.print_raise_again(x, SOME "update1") )

    | update (head::tail, start, value) =
      ( let val l = Word.fromInt(Word8Array.length value)
	   fun loop([],  old_i, new_i) =
	         if (new_i <> l) then illegal_access "update longer than T"
		 else ()
	     | loop((head as {data,start,length})::tail, old_i, new_i) =
		 if (new_i = l) then () 
		 else if old_i < length
		 then( Word8Array.update(data, Word.toInt(old_i+start), 
					 Word8Array.sub(value,
							Word.toInt new_i));
		       loop(head::tail,  old_i+0w1, new_i+0w1) )
		 else loop(tail, 0w0, new_i)
	       
	   fun skip_loop( [], 0w0 ) = ([], 0w0)
	     | skip_loop( [], _) = illegal_access "update starts past T"
	     | skip_loop((head as {data, start, length}) :: tail, count) =
		if count >= length 
		then skip_loop(tail, count - length) 
		else (head::tail, count)	 
           val (list, pos) = skip_loop(head::tail, start)
        in loop(list, pos, 0w0)
	end )	 handle x =>Trace.print_raise_again(x, SOME "update2") 	 
	  		 

 
(*

  fun update (Empty, _, _) = illegal_access "updating empty T"
    | update (Single array, start, value) =
       let fun loop (NONE, _) = ()
	     | loop (_, NONE) = illegal_access "update longer than T"
	     | loop (SOME (v, rest), SOME updatable) =
	        loop (WAF.next rest, WAF.update (updatable, v))
       in loop (WAF.next (Word_Array.to8 value),
		WAF.write (WAF.seek (array, start)))
       end
    | update (array as (Multiple (head, tail)), start, value) =
       let fun loop (NONE, (_, _)) = ()
	     | loop (_, (NONE, [])) = illegal_access "update longer than T"
	     | loop (v, (NONE, head :: tail)) =
	        loop (v, (WAF.write head, tail))
	     | loop (SOME (v, rest), (SOME updatable, list)) =
	        loop (WAF.next rest,
		      (WAF.update (updatable, v), list))
	   fun skip_loop ([], 0w0) = (NONE, [])
	     | skip_loop ([], _) = illegal_access "update starts past T"
	     | skip_loop (head :: tail, count) =
	        let val head_length = WAF.length head
		in if count >= head_length then
		    skip_loop (tail, count - head_length)
		   else (NONE, WAF.seek (head, count) :: tail)
		end
       in loop (WAF.next (Word_Array.to8 value),
		skip_loop (head :: tail, start))
       end
*)


  fun join (a1, a2) = a1@a2



  fun split_single (array as {data, start, length}, position:word, _) =
       ([{data=data, start=start, length=position}],
	[{data=data, start=start+position, length = length-position}])

(*
  fun split_single (array, position, length) =
       (Single (WAR.seek (array, length - position)),
	Single (WAF.seek (array, position)))
*)

  fun split_multiple (first, second :: rest, 0w0) =
       ([], first::second :: rest)
    | split_multiple (head as {data, start, length}, [], position) =
       if (position >= length) 
       then illegal_access "split position longer than T"
       else
             ([{data=data, start=start, length=position}],
	      [{data=data, start=start+position, length = length-position}])
    | split_multiple (first as {data, start, length}, 
		      second :: rest, position) =
       if position < length then
	   ([{data=data, start=start,length=position}],
	    {data=data, start=start+position,length=length-position}::
	     second :: rest)
       else
	   let val (first_half, second_half) =
	             split_multiple (second, rest, position - length)
	   in (join ([first], first_half), second_half)
	   end
      
(*
  fun split_multiple (first, second :: rest, 0w0) =
       (Empty, Multiple (first, second :: rest))
    | split_multiple (head, [], position) =
       (Single (WAR.seek (head, WAF.length head - position)),
	Single (WAF.seek (head, position)))
    | split_multiple (first, second :: rest, position) =
       let val length = WAF.length first
       in if position < length then
	   (Single (WAR.seek (first, length - position)),
	    Multiple (WAF.seek (first, position), second :: rest))
	  else
	   let val (first_half, second_half) =
	             split_multiple (second, rest, position - length)
	   in (join (Single first, first_half), second_half)
	   end
       end
*)

  fun split (array, position) =
     ((  let val length = size array
       in if position < 0w0 then
	   illegal_access "negative split position"
          else if position > length then
	   illegal_access "split position beyond end of T"
          else if position = length then (array, [])
          else if position = 0w0 then ([], array)
          else
	   case array of
	      [] => illegal_access "splitting empty T"
	    | [a as {...}] => split_single (a, position, length)
	    | head::tail => split_multiple (head, tail, position)
       end )  handle x =>Trace.print_raise_again(x, SOME "split") )

(*
	2.	function fold
*)

 fun fold ([], f, init) = init
    | fold ((array as{data, start, length})::[], f, init) =
       ((f (array, init) ) 
	handle x =>Trace.print_raise_again(x, SOME "fold/user function") )
    | fold (first::rest, f, init) =
       ((let fun fold_forward ([], result) = result
	     | fold_forward ((head as{data, start, length}) :: tail, result) =
	        fold_forward (tail, f (head, result))
       in fold_forward (first :: rest, init)
       end )
	   handle x =>Trace.print_raise_again(x, SOME "fold") )

(*
  fun fold (Empty, f, init) = init
    | fold (Single array, f, init) = f (Word_Array.from8 array, init)
    | fold (Multiple (first, rest), f, init) =
       let fun fold_forward ([], result) = result
	     | fold_forward (head :: tail, result) =
	        fold_forward (tail, f (Word_Array.from8 head, result))
       in fold_forward (first :: rest, init)
       end
*)
(*
	3.	functions makestring and makestring_max
*)

  fun makestring_max ([], _) = ""
    | makestring_max ([{data,start,length}], count) =
     ((  let fun loop (str, index) =
	       if index = count 
	       then str
	       else loop(str^ "." ^
			 Word8.toString(Word8Array.sub(data, 
						       Word.toInt(start+index))),
			 index+0w1)
       in  if count > length 
	   then illegal_access "count is larger than length"
	   else loop("", 0w0)
       end)
	   handle x =>Trace.print_raise_again(x, SOME "makestring_max1") )
    | makestring_max (head::tail, count) =
       ((let fun loop([], str, count) = str
	     | loop(head::tail, str, count) =
	       loop(tail, str^ "/" ^
		    makestring_max([head], size([head])), count - size([head]))
	in loop(head::tail, "", count)
	end )
	   handle x =>Trace.print_raise_again(x, SOME "makestring_max2") )

  fun makestring array = makestring_max (array, size array)

 end (* struct *)









