(*

	FoxNet: The Fox Project's Communication Protocol Implementation Effort
	Edoardo Biagioni <Edoardo.Biagioni@cs.cmu.edu>
	Ken Cline (cline+@cs.cmu.edu)
	Brian Milnes (Brian.Milnes@cs.cmu.edu)
	Nick Haines  (Nick.Haines@cs.cmu.edu)
	Fox Project
	School of Computer Science
	Carnegie Mellon University
	Pittsburgh, Pa 15139-3891

		i.	Abstract

	A setup for constructing and comparing dynamic arrays for testing
	purposes.

		ii.	Table of Contents

	i.	Abstract
	ii.	Table of Contents
	iii.	RCS Log
	1.	functor Compare_Dyn_Array
	2.	types difference and data
	3.	function create
	4.	function compare
	5.	function makestring


		iii.	RCS Log
	
$Log: compareword.fun,v $
Revision 1.1  1995/09/18  19:33:43  esb
Initial revision


		1.	functor Compare_Word_Array
*)

functor Compare_Word_Array (structure Debug: DEBUG
			    structure V: VENDOR): COMPARE_ARRAY  =
 struct 
  structure Trace = Trace (structure V = V
			   val debug_level = NONE
			   val module_name = "comparedyn.fun"
			   val makestring = fn _ => NONE)

(*
		2.	types difference and data
*)

  datatype difference = Data_Difference of int * FoxWord8.word * FoxWord8.word 
                      | Data_Size of int * int

  type data = Word_Array.T

(*
		3.	function create

	Create a Word_Array.T with the given header and
	data bytes filled out.
*)

  fun create list =
       Word_Array.W8.new (fn [] => NONE | (f :: r) => SOME (f, r)) list

(*
		4.	function compare
*)
 
  fun compare this that =
       let val this_len = Word_Array.W8.length this
           val that_len = Word_Array.W8.length that
	   fun loop (NONE, _) = []
	     | loop (_, NONE) = []
             | loop (SOME (this_head, this_rest),
		     SOME (that_head, that_rest)) = 
                if this_head <> that_head then
	         Data_Difference (this_len + 1 -
				  Word_Array.W8.length this_rest,
				  this_head, that_head) ::
	         loop (Word_Array.W8.next this_rest,
		       Word_Array.W8.next that_rest)
                else loop (Word_Array.W8.next this_rest,
			   Word_Array.W8.next that_rest)
      in if this_len <> that_len then 
          Data_Size (this_len, that_len) ::
	  loop (Word_Array.W8.next this, Word_Array.W8.next that)
         else loop (Word_Array.W8.next this, Word_Array.W8.next that)
      end

(*
		5.	function makestring
*)

  fun makestring [] = ""
    | makestring (head :: rest) = 
       (case head of
           Data_Difference (i, b1, b2) =>
	    "Data_Difference " ^ V.Integer.makestring i ^ " " ^
	    FoxMakestring.word8 b1 ^ " " ^ FoxMakestring.word8 b2 ^
	    makestring rest
	 | Data_Size (i1, i2) =>
	    "Data_Size " ^
	    V.Integer.makestring i1 ^ " " ^ V.Integer.makestring i2 ^
	    makestring rest)

 end (* struct *)



